/*
 * colorspec.c 
 *
 * RGB color image of a spectrum
 * Version 1.0 : 22 July 2004 
 * J. Kielkopf
 *
 * Compile:
 *
 *   gcc -lm colorspec.c
 *
 *
 * Usage:
 *
 *   colorspec flux_file ppm_file
 *
 *     flux_file  --  Input flux file
 *     ppm_file  --  Output image  file
 *
 */
 
 /* Include */

#include <stdio.h>              /* Standard io library */
#include <math.h>               /* Math library */
#include <signal.h>             /* Error handling */
#include <stdlib.h>             /* Number conversion */
#include <string.h>             /* String functions */
#include <time.h>               /* Date and time */
#include <sys/types.h>          /* Typedefs */
#include <sys/stat.h>           /* Structure returned by stat */
#include <fcntl.h>              /* Flags for read and write */


/* Global constants */

#define NW 800                  /* Image width */
#define NH 50                   /* Image height */
#define ND 255                  /* Color depth */
#define NL 800000               /* Maximum number of samples in table */


/* Function prototypes */

int readtf(void);               /* Read flux table file */
void colorimg(void);            /* Calculate the color image */
void writeppm(void);            /* Write the image file */
void wl2rgb(double x);          /* Convert wl=x to rgb triplet */
double fa(double x, double dx);  /* Emergent flux at x in bandwidth dx */


/* Data structures */

struct colorimg                 /* Structure for rgb values */
 {
  int r;
  int g;
  int b;
 };

struct fluxlist                /* Structure for a spectrum table */
 {
  double wl;
  double flux;
 }; 


struct colorimg icv[NW];           /* Color value array for the spectrum */
struct fluxlist  spectrum[NL];     /* Digital spectrum */


/* Global data file parameters */

FILE *flux_file;                     /* File pointer for line file */
FILE *ppm_file;                     /* File pointer for ppm file */
char flux_name[64];                  /* File name for line file */
char ppm_name[64];                  /* File name for ppm file */


/* Global data with default values */

int nl;                            /* How many flux entries */
double pi=3.141592654;             /* PI */


/* Global variables */

double r,g,b;                      /* Buffer rgb values */
double fluxmax;                    /* Maximum flux 3800 to 7800 A */



/* Main program */

int main(int argc,char *argv[])
{ 
  int flag;
  if(argc != 3)
  {  
    printf("Usage: colorspec flux_name ppm_name\n");
    exit(1);
  }
  
  /* Convert argument strings to values */
  
  sscanf(argv[1], "%s",  flux_name);
  sscanf(argv[2], "%s",  ppm_name);

  printf(" \n");
  printf(" Flux file name: %s \n", flux_name);
  printf(" Image ppm file name: %s \n", ppm_name);

  
  /* Read the flux table file */

  flag=readtf();
  if(flag == 1)
  {
    printf(" Unable to read the flux file.\n");
    exit(1);
  }

  printf(" Read %d lines from the file.\n",nl);
  printf(" Maximum flux from 3800 to 7800 A was %g.\n",fluxmax);
  
  colorimg();
  writeppm();
  
  exit(0);
}


/* Read the line list table */
 
int readtf(void)
{
  char buf[101];
  int i;
  double wl,flux;
    
  /* Open the file and test for success */

  flux_file = fopen(flux_name,"r");   
  
  if (flux_file == NULL ) 
  { 
    printf("\n  Unable to open %s \n", flux_name);
    return 1;
  }
  
  /* Read the entries and find the maximum flux */
  
  fluxmax=0.;
  
  for( i=0 ;  (fgets(buf,100,flux_file) != NULL) & (i < NL) ; i++ )
  {
    sscanf(buf,"%lf %lf",&wl, &flux);    
    spectrum[i].wl = wl;
    spectrum[i].flux = flux;
    if ( (wl >= 3800.) & (wl <= 7800.) )
    {
      if(flux > fluxmax)
      {
        fluxmax = flux;
      }
    }
  }
  nl=i;

  fclose(flux_file);
  return 0;
}


/* Create a color image m wide by n high */

void colorimg(void)
{

  int i,j,m,n,mx;
  double amx,gam,dw,a,wl;
  int rr,bb,gg;

  m=NW;
  n=NH;
  mx=ND;
  amx = (double) mx;
  gam=0.8;
  dw=4000./( (double) (m-1) );
  
  /* Find color value, icv, of each pixel */
  
  for( i=0 ; i < m ; i++ )
  {
    wl=3800. + dw * ((double) i);
    wl2rgb(wl);
    a=fa(wl,dw);
    a=a/fluxmax;
 
    icv[i].r = (int) ( amx * pow( a*r , gam ) );
    icv[i].g = (int) ( amx * pow( a*g , gam ) );
    icv[i].b = (int) ( amx * pow( a*b , gam ) ); 
  }
} 

/* Write a ppm color image file for the array icv */   

void writeppm(void)
{

  int i,j,m,n,mx;

  m=NW;
  n=NH;
  mx=ND;

  /* Open ppm file for m wide by n high image normalized to color depth mx */

  ppm_file = fopen(ppm_name,"w");

  fprintf(ppm_file,"P3        \n");
  fprintf(ppm_file,"# spectrum\n");
  fprintf(ppm_file,"%d %d\n",m,n);
  fprintf(ppm_file,"%d\n",mx);
   
  /* Write the ppm image */

  for( j=0 ; j < n ; j++ )
  {
    for( i=0 ; i < m ; i++ )
    {
      fprintf(ppm_file,"%d %d %d\n",icv[i].r,icv[i].g,icv[i].b); 
    }
  }
  close(ppm_file);   
}


/*  Calculate color values rgb for a visible wavelength wl in Angstroms  */
/*  Based on a program by Dan Bruton (astro@tamu.edu) found at */
/*    http://www.isc.tamu.edu/~astro/color.html */
/*  Modified by John Kielkopf (kielkopf@louisville.edu) */
/*    to improve match of color rendition to visual response */

void wl2rgb(double wl)
{
  double rhi,rlo;    

  r=0.;
  g=0.;
  b=0.;
  if((wl >= 3800.) & (wl <= 7800.))
  {
    rhi=1.0/(1.+exp(((5400.-wl)/150.)));
    rhi=rhi/(1.+exp(((wl-6700.)/200.)));
    rlo=0.4/(1.+exp(((wl-4400.)/300.)));
    rlo=rlo/(1.+exp(((3900.-wl)/200.)));
    r=rhi+rlo;
    g=1.0/(1.+exp(((wl-6000.)/200.)));
    g=g/(1.+exp(((4800.-wl)/150.)));
    g=g*1.02;
    b=1.0/(1.+exp(((wl-5000.)/150.)));
    b=b/(1.+exp(((4000.-wl)/200.)));
    b=b*1.0;
  }

  return;
}



/*  Average relative emergent flux (0 to 1) */
/*    in a bandwidth dw centered on wl */


double fa(double wl, double dw)
{
  double sum,wllo,wlhi;
  int i,j;
  sum=0.;
  i=0;
  
  wllo=wl-dw/2.;
  wlhi=wl+dw/2.; 
  
  /* No flux if the bandwidth does not overlap the flux table */
  
  if(wlhi < spectrum[0].wl)
  {
    return 0.;
  } 
  if(wllo > spectrum[(nl - 1)].wl)
  {
    return 0.;
  }   
  
  /* Identify and average entries in the table that are in the band */
  /* The table should be ordered low to high */
    
  for( j=0 ; j < nl ; j++ )
  { 
    wl=spectrum[j].wl;
    
    /* Start with the last element below the band */
    
    if( wl < wllo )
    {
      sum=spectrum[j].flux;
      i=1;
    } 
    
    /* Add an element in the band */
       
    if( (wl >= wllo) & (wl <= wlhi) )
    {
      sum=sum+spectrum[j].flux;
      i++;
    }    

    /* Stop with the first element above the band */
    
    if( wl > wlhi )
    {
      sum=sum+spectrum[j].flux;
      i++;
      break;
    } 
  }
  
  /* Calculate an average */
  
  if(i>1)
  {
    sum = sum / ( (double) i );
  }
    
  return sum;
}
