/*
 * solar.c 
 *
 * RGB Solar spectrum from a line list
 * Version 1.2 : 22 July 2004 
 * J. Kielkopf
 *
 * Compile:
 *
 *   gcc -lm solar.c 
 *
 *
 * Usage:
 *
 *   solar tf_file sf_file sw
 *
 *     tf_file  --  Input line list
 *     sf_file  --  Output ppm image
 *     sw       --  HWHM of typical spectral line
 *
 */
 
 /* Include */

#include <stdio.h>              /* Standard io library */
#include <math.h>               /* Math library */
#include <signal.h>             /* Error handling */
#include <stdlib.h>             /* Number conversion */
#include <string.h>             /* String functions */
#include <time.h>               /* Date and time */
#include <sys/types.h>          /* Typedefs */
#include <sys/stat.h>           /* Structure returned by stat */
#include <fcntl.h>              /* Flags for read and write */


/* Global constants */

#define NW 800                  /* Image width */
#define NH 50                   /* Image height */
#define ND 255                  /* Color depth */
#define NL 30000                /* Maximum number of lines in table */


/* Function prototypes */

int readtf(void);               /* Read line list table file */
void colorimg(void);            /* Calculate the color image */
void writeppm(void);            /* Write the image file */
void wl2rgb(double x);          /* Convert wl=x to rgb triplet */
double spectrum(double x, double dx);   /* Pixel luminosity at wl=x for dw=dx*/
double fa(double x);            /* Emergent flux at wl=x */


/* Data structures */

struct colorimg                 /* Structure for rgb values */
 {
  int r;
  int g;
  int b;
 };

struct linelist                /* Structure for a table of spectral lines */
 {
  double wl;
  double ew;
 }; 


struct colorimg icv[NW];           /* Color value array for the spectrum */
struct linelist line[NL];          /* Line list for the spectrum */


/* Global data file parameters */

FILE *tf_file;                     /* File pointer for line file */
FILE *sf_file;                     /* File pointer for ppm file */
char tf_name[64];                  /* File name for line file */
char sf_name[64];                  /* File name for ppm file */


/* Global data with default values */

int nl;                            /* How many line list entries */
double pi=3.141592654;             /* PI */
double sw=0.3;                     /* HWHM of spectral line */


/* Global variables */

double r,g,b;                      /* Buffer rgb values */



/* Main program */

int main(int argc,char *argv[])
{ 
  int flag;
  if(argc != 4)
  {  
    printf("Usage: solar tf_name sf_name sw\n");
    exit(1);
  }
  
  /* Convert argument strings to values */
  
  sscanf(argv[1], "%s",  tf_name);
  sscanf(argv[2], "%s",  sf_name);
  sscanf(argv[3], "%lf",  &sw);

  printf(" \n");
  printf(" Line list file: %s \n", tf_name);
  printf(" Spectrum image: %s \n", sf_name);
  printf(" HWHM (A): %lf \n", sw);

  
  /* Test validity of input */
  
  if(sw <= 0.)
  {
    sw=0.3;
  }
  

  /* Read the line list table file */

  flag=readtf();
  if(flag == 1)
  {
    printf(" Unable to read the line list file.\n");
    exit(1);
  }

  printf(" Read %d lines from the file.\n",nl);
  
  colorimg();
  writeppm();
  
  exit(0);
}


/* Read the line list table */
 
int readtf(void)
{
  char buf[101];
  int flag,i;
  double wl,ew;
    
  /* Open the file and test for success */

  tf_file = fopen(tf_name,"r");   
  
  if (tf_file == NULL ) 
  { 
    printf("\n  Unable to open %s \n", tf_name);
    return 1;
  }
  
  /* Read the entries */
  
  for( i=0 ;  (fgets(buf,100,tf_file) != NULL) & (i < NL) ; i++ )
  {
    sscanf(buf,"%lf %lf",&wl, &ew);    
    line[i].wl = wl;
    line[i].ew = ew/1000.;
  }
  nl=i;

  fclose(tf_file);
  return 0;
}


/* Create a color image m wide by n high */

void colorimg(void)
{

  int i,j,m,n,mx;
  double amx,gam,dw,a,wl;
  int rr,bb,gg;

  m=NW;
  n=NH;
  mx=ND;
  amx = (double) mx;
  gam=0.8;
  dw=4000./( (double) (m-1) );
  
  /* Find color value, icv, of each pixel */
  
  for( i=0 ; i < m ; i++ )
  {
    wl=3800. + dw * ((double) i);
    wl2rgb(wl);
    a=spectrum(wl,dw);
 
    icv[i].r = (int) ( amx * pow( a*r , gam ) );
    icv[i].g = (int) ( amx * pow( a*g , gam ) );
    icv[i].b = (int) ( amx * pow( a*b , gam ) ); 
  }
} 

/* Write a ppm color image file for the array icv */   

void writeppm(void)
{

  int i,j,m,n,mx;

  m=NW;
  n=NH;
  mx=ND;

  /* Open ppm file for m wide by n high image normalized to color depth mx */

  sf_file = fopen(sf_name,"w");

  fprintf(sf_file,"P3        \n");
  fprintf(sf_file,"# spectrum\n");
  fprintf(sf_file,"%d %d\n",m,n);
  fprintf(sf_file,"%d\n",mx);
   
  /* Write the ppm image */

  for( j=0 ; j < n ; j++ )
  {
    for( i=0 ; i < m ; i++ )
    {
      fprintf(sf_file,"%d %d %d\n",icv[i].r,icv[i].g,icv[i].b); 
    }
  }
  close(sf_file);   
}


/*  Calculate color values rgb for a visible wavelength wl in Angstroms  */
/*  Based on a program by Dan Bruton (astro@tamu.edu) found at */
/*    http://www.isc.tamu.edu/~astro/color.html */
/*  Modified by John Kielkopf (kielkopf@louisville.edu) */
/*    to improve match of color rendition to visual response */

void wl2rgb(double wl)
{
  double rhi,rlo;    

  r=0.;
  g=0.;
  b=0.;
  if((wl >= 3800.) & (wl <= 7800.))
  {
    rhi=1.0/(1.+exp(((5400.-wl)/150.)));
    rhi=rhi/(1.+exp(((wl-6700.)/200.)));
    rlo=0.4/(1.+exp(((wl-4400.)/300.)));
    rlo=rlo/(1.+exp(((3900.-wl)/200.)));
    r=rhi+rlo;
    g=1.0/(1.+exp(((wl-6000.)/200.)));
    g=g/(1.+exp(((4800.-wl)/150.)));
    g=g*1.02;
    b=1.0/(1.+exp(((wl-5000.)/150.)));
    b=b/(1.+exp(((4000.-wl)/200.)));
    b=b*1.0;
  }

  return;
}


/*  Integrate spectrum over a pixel at wl in Angstroms */
/*  Each pixel covers wl \pm dw/2 */
/*  This model will have a correct line shape for the optically thin case */
/*    but the equivalent width is measured in the absorption spectrum */
/*    and we apply it here to the absorption coefficient without correction */
/*    for a curve of growth */

/*  Center wavelength wl, hwhm sw, and equivalent width ew are in Angstroms */
/*  Visible spectrum from 3800 to 7800 Anstroms covered by nw-1 pixels */

double spectrum(double wl,double dw)
{
  double wllo,wlhi,a,h,sum,x;
  int j,n;

  wllo=wl-dw/2.;
  wlhi=wl+dw/2.;

  /* Integration from wllo to wlhi */
  /* Subdivide the intrinsic line hwhm into elements sw/8 wide */
  /* Integrate over n elements with the extended trapezoidal rule */

  h=sw/8.;
  n= (int) (dw/h);
  if(n < 1)
  {
    n=1;
  }
   
  if(n == 1)
  {
    sum=fa(wl);
  }
   
  if( n >= 2) 
  {
    sum=0.5*(fa(wllo)+fa(wlhi));
  }
  
  if( n > 2)
  {
    x=wllo;
    for( j=2 ; j < (n -1) ; j++ )
    {
      x=x+h;
      sum=sum+fa(x);
    }
  }
  
  a=sum/((double) n);
  return a;
}


/*  Relative emergent flux (0 to 1) at wavelength wl (Angstroms) */
/*  The absorption at wl is a sum over all lines */
/*  Absorption coefficient for each line has hwhm sw */
/*  Intrinsic Lorentzian line profile area normalized to ew */

double fa(double wl)
{
  double flux,wl0,ew,a;
  int j;
  flux=1.;
  for( j=0 ; j < nl ; j++ )
  { 
    wl0=line[j].wl;
    ew=line[j].ew;
    a=(ew*sw/pi)/(sw*sw+(wl-wl0)*(wl-wl0));
    flux=flux*exp(-a);
  }
  return flux;
}
