      program solar
     
!     RGB Solar spectrum from a line list
!     Version 1.2 : 20 July 2004 
!     J. Kielkopf

!     Usage:
!
!     Read a list of lines with wavelength and equivalent width
!     Generate an rgb ppm image file for a visible solar spectrum
!     Convert ppm file to other formats with Gimp or ImageMagick

!     Input data:
!
!     Moore_List.txt contains Charlotte Moore's NBS 61 Solar Spectrum
!     Sorted versions of this list contain lines of selected elements
!     Input is free form --
!       Column one:  wavelength in A
!       Column two:  equivalent width in mA

!     Output image format:
!
!     This program produces a plain text ppm color image
!     Select width and height by changing the parameters below
!     Visible wavelengths from 3800 to 7800 A are mapped to this image

      
      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw
      common /fnames/tfname,sfname
      character tfname*32,sfname*32

      call control
      call readtf     
      call colorimg
      call writeppm
      
      end
 
 
      subroutine control
      
!     Obtain the file names and parameters
!       tfname -- line list input file name
!       sfname -- spectrum output file name

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw     
      common /fnames/tfname,sfname
      character tfname*32,sfname*32

      write(*,'(a,$)') 'Line list input file: '      
      read(*,'(a)') tfname
      write(*,'(a,$)') 'Spectrum output ppm file: '
      read(*,'(a)') sfname
      write(*,'(a,$)') 'Solar line hwhm (A) [0.3]: '
      read(*,*) sw
      if(sw.le.(0.d0)) then
        sw=0.3
      endif
      
      end


      subroutine readtf

!     Read line list

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw      
      common /fnames/tfname,sfname
      character tfname*32,sfname*32

      open(unit=21,file=tfname,status='old')

!     Read a line list
!     Wavelength in Angstroms
!     Equivalent width in milliAngstroms in the table is converted to Angstroms

      nlines=0
      do i=1,nl
         read(21,*,end=20) alambda(i),ewidth(i)
         ewidth(i)=ewidth(i)/1000.d0
         write(*,'(1x,i5,1x,d12.6,1x,d12.6)') i,alambda(i),ewidth(i)
         nlines=i
      enddo
 20   close(21)

      end

            

      subroutine colorimg
      
!     Make an integer color image of a visible spectrum

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw      
      common /fnames/tfname,sfname
      character tfname*32,sfname*32
      
!     Define image  - width, height, depth, gamma

      m=nw
      n=nh
      mx=nd
      gam=0.8
      dw=(4000.d0)/dfloat(nw-1)

!     Find color value, icv, of each pixel 

      do i=1,m
       wl=3800.d0 + dw*dfloat(i-1)
       call wl2rgb(wl,r,g,b)
       call spectrum(wl,dw,a)
       do j=1,n
        rmax=1.d0
        icv(i,j,1)=int(real(mx)*(a*r/rmax)**gam)
        icv(i,j,2)=int(real(mx)*(a*g/rmax)**gam)
        icv(i,j,3)=int(real(mx)*(a*b/rmax)**gam)
       enddo
      enddo
      end
      
      

      subroutine writeppm

!     Write a ppm color image file for the array icv

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw      
      common /fnames/tfname,sfname
      character tfname*32,sfname*32      

      m=nw
      n=nh
      mx=nd     

!     Output ppm file normalized to color depth mx

      open(unit=20,file=sfname,status='unknown')

      write(20,'(a10)') 'P3        '
      write(20,'(a10)') '# spectrum'
      write(20,*) m,n
      write(20,*) mx
      do j=1,n
       do i=1,m
        write(20,*) icv(i,j,1),icv(i,j,2),icv(i,j,3)
       enddo
      enddo
      close(20)
      end 


      subroutine wl2rgb(wl,r,g,b)
      implicit double precision (a-h,o-z)
      
!     Calculate color values rgb for a visible wavelength wl in Angstroms
!     Based on a program by Dan Bruton (astro@tamu.edu) found at
!       http://www.isc.tamu.edu/~astro/color.html
!     Modified by John Kielkopf (kielkopf@louisville.edu)
!       to improve match of color rendition to visual response

      r=0.d0
      g=0.d0
      b=0.d0
      if((wl.ge.(3800.d0)).and.(wl.le.(7800.d0))) then
       rhi=1.0d0/(1.d0+dexp(((5400.d0-wl)/150.d0)))
       rhi=rhi/(1.d0+dexp(((wl-6700.d0)/200.d0)))
       rlo=0.4d0/(1.d0+dexp(((wl-4400.d0)/300.d0)))
       rlo=rlo/(1.d0+dexp(((3900.d0-wl)/200.d0)))
       r=rhi+rlo
       g=1.0d0/(1.d0+dexp(((wl-6000.d0)/200.d0)))
       g=g/(1.d0+dexp(((4800.d0-wl)/150.d0)))
       g=g*1.02d0
       b=1.0d0/(1.d0+dexp(((wl-5000.d0)/150.d0)))
       b=b/(1.d0+dexp(((4000.d0-wl)/200.d0)))
       b=b*1.0d0
      endif

      return
      end
 
       
      subroutine spectrum(wl,dw,a)

!     Calculate the integrated spectrum over a pixel at wl in Angstroms
!     Each pixel integrates over wl \pm dw/2
!     This model will have a correct line shape for the optically thin case
!       but the equivalent width is measured in the absorption spectrum
!       and we apply it here to the absorption coefficient without correction
!       for a curve of growth

!     Center wavelength wl, hwhm sw, and equivalent width ew are in Angstroms
!     Visible spectrum from 3800 to 7800 Anstroms covered by nw-1 pixels

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw
      common /fnames/tfname,sfname
      character tfname*32,sfname*32
      wllo=wl-dw/2.d0
      wlhi=wl+dw/2.d0

!     Integration from wllo to wlhi
!     Subdivide the intrinsic line hwhm into elements sw/8 wide
!     Integrate over n elements with the extended trapezoidal rule

      h=sw/8.d0
      n=dw/h
      if(n.lt.1) then
        n=1
      endif  
      if(n.eq.1) then
        sum=fa(wl)
      endif
      if(n.ge.2) then
        sum=0.5d0*(fa(wllo)+fa(wlhi))
      endif
      if(n.gt.2) then
        x=wllo
        do i=2,n-1
          x=x+h
          sum=sum+fa(x)
        enddo
      endif     
      a=sum/dfloat(n)
      return
      end
      
            
      real*8 function fa(wl)
      
!     Relative emergent flux fa (0 to 1) at wavelength wl (Angstroms)
!     Calculate the absorption at wl by summing all lines
!     Absorption coefficient for each line has hwhm sw
!       and intrinsic Lorentzian profile area normalized to ew

      parameter (nw=800)
      parameter (nh=50)
      parameter (nd=255)
      parameter (nl=30000)
      implicit real*8 (a-h,o-z)
      common /color/icv(nw,nh,3)
      common /lines/alambda(nl),ewidth(nl),nlines
      common /shape/sw
      common /fnames/tfname,sfname
      character tfname*32,sfname*32
      pi=3.141592654d0
      fa=1.d0
      do i=1,nlines
        wl0=alambda(i)
        ew=ewidth(i)
        a=(ew*sw/pi)/(sw*sw+(wl-wl0)*(wl-wl0))
        fa=fa*dexp(-a)
      enddo
      
      return
      end
